open Js_of_ocaml
open Js_of_ocaml_lwt

module Option = struct

  let iter opt f = begin match opt with
    | None -> ()
    | Some x -> f x
  end

end

(** Event generated once the library has created the svg image *)
let (generated_svg_event:(Dom_svg.svgElement Js.t * Js.js_string Js.t) Dom.Event.typ)
  = Dom_html.Event.make "generated_svg"

class type viz = object
  (* PNG output *)
  method renderImageElement: Js.js_string Js.t -> (Dom_html.imageElement Js.t, unit) Promise.promise Js.meth
  (* SVG output *)
  method renderSVGElement: Js.js_string Js.t -> (Dom_svg.svgElement Js.t, unit) Promise.promise Js.meth
end

let (viz: viz Js.t Js.constr)
  = Js.Unsafe.global##._Viz

(* Save the generated image *)
let on_result_png image = begin
  let a = Dom_html.createA Dom_html.document in
  a##.href := image##.src;
  Js.Unsafe.set a (Js.string "download") (Js.string "output.png");
  Dom.appendChild Dom_html.document##.body a;
  a##click;
  Dom.removeChild Dom_html.document##.body a
end

let on_result_svg
  : Dom_html.element Js.t -> (Dom.node Js.t) -> unit
  = fun svg_display image -> begin
      let childNodes = svg_display##.childNodes in
      for i = 0 to childNodes##.length do
        Js.Opt.iter (childNodes##item i) (fun elem -> ignore @@ svg_display##removeChild elem)
      done;
      Dom.appendChild svg_display image;
      ()
    end

let download data (filename : Js.js_string Js.t) (mime: Js.js_string Js.t)
  = Js.Unsafe.(
      fun_call (js_expr "download")
        [| inject data
         ; inject filename
         ; inject mime |])

(* Update the Export PNG button and change the output image. *)
let update_button png_output _gv image = begin
  png_output##.onclick := Dom_html.handler (fun _->
      (*download image*)
      let _ = download image (Js.string "output.png") (Js.string "text/plain") in
      Js._false
    );
end

let change_event input _ev _ = begin
  let text: Js.js_string Js.t = input##.value in
  if text##.length != 0 then (
    let gv = Formatter.convert text in
    let v = new%js viz in
    let _ = Promise.then_map
        (try v##renderSVGElement gv with
         | _ -> Promise.reject ())
        ~on_fulfilled:(fun image ->
            Dispatch.call generated_svg_event (image, gv) )
    in ()
  );
  Lwt.return_unit
end

let load () =
  let svg_display = Dom_html.getElementById_exn "svg"
  and output_button = Dom_html.getElementById_exn "png_output"
  and dot_button = Dom_html.getElementById_exn "dot_output"
  and editor = Dom_html.getElementById_exn "editor" in

  ignore @@ Dispatch.register generated_svg_event (fun (image, input) ->

      on_result_svg svg_display (image :> Dom.node Js.t);

      (* Convert the svg in png *)
      CanvasTool.generate_png image (fun image ->
          update_button output_button input image;
        );

      (* Download the dot file *)
      dot_button##.onclick := Dom_html.handler (fun _ev ->
          let _ = download input (Js.string "output.dot") (Js.string "text/plain") in
          Js._false
        );
    );

  Js.Opt.iter (Dom_html.CoerceTo.textarea editor) (fun input ->
      Lwt_js_events.async (fun () -> (Lwt_js_events.limited_loop Lwt_js_events.change ~elapsed_time:0.1 editor (change_event input) ));
      Lwt_js_events.async (fun () -> (Lwt_js_events.limited_loop  Lwt_js_events.input ~elapsed_time:0.1 editor (change_event input) ));
      Lwt.ignore_result (change_event input () ());
    );

  (** Handle the separation window *)
  Option.iter (Dom_html.getElementById_opt "btn_window") (fun button ->
      button##.onclick := Dom_html.handler (fun _ ->

          let opt_window2 = Dom_html.window##open_
              (Js.string "")
              (Js.string "_image")
              Js.Opt.empty in

          Js.Opt.iter opt_window2 (fun window2 ->

              let callback
                : Dom_svg.svgElement Js.t * Js.js_string Js.t -> unit
                = fun (image, _gv)  -> begin
                    let node2 = image##cloneNode Js._true in
                    let () = on_result_svg window2##.document##.body (node2 :> Dom.node Js.t) in
                    ()
                  end in
              let handler = Dispatch.register generated_svg_event callback in

              (* Remove the update event when the window is closed *)
              window2##.onunload := Dom_html.handler (fun _ev ->
                  Dispatch.remove generated_svg_event handler;
                  Js._false
                );

              (* Copy the current display to the new window *)
              let childNodes = svg_display##.childNodes in
              for i = 0 to childNodes##.length do
                Js.Opt.iter (childNodes##item i) (fun elem ->
                    Dom.appendChild window2##.document##.body (elem##cloneNode Js._true)
                  ) done;

            );

          Js._false
        )
    );
  Lwt.return Js._false

let _ = Lwt.bind (Lwt_js_events.onload ()) (fun _ -> load())

let _ =
  Js.export "generator"
    (object%js
      method load = load ()
    end)
